import logging
import subprocess

from whoare.base import Domain, subclasses
from whoare.exceptions import ZoneNotFoundError, WhoIsCommandError

logger = logging.getLogger(__name__)


class WhoAre:

    def __init__(self):
        self.child = None  # the object for the zone
        self.domain = None  # Domain Object
        self.registrant = None  # Registrant Objects
        self.dnss = []  # all DNSs objects

    def as_dict(self):
        """ build a nice dict """
        res = {
                "domain": {
                    "base_name": self.domain.base_name,
                    "zone": self.domain.zone,
                    "is_free": self.domain.is_free
                    }
                }
        if not self.domain.is_free:
            res["domain"]["registered"] = self.domain.registered
            res["domain"]["changed"] = self.domain.changed
            res["domain"]["expire"] = self.domain.expire
            res["registrant"] = {
                "name": self.registrant.name,
                "legal_uid": self.registrant.legal_uid,
                "created": self.registrant.created,
                "changed": self.registrant.changed
                }

            res["dnss"] = [dns.name for dns in self.dnss]
            
        return res

    def load(self, domain, host=None, mock_from_txt_file=None):
        """ load domain data. 
                domain is DOMAIN.ZONE (never use subdomain like www or others)
                host could be "whois.nic.ar" of rargentina (optional) 
            Return a dict with parsed data and fill class properties 
                mock_from_txt_file could be used to a path with exact whois results """
        
        logger.info(f'Load {domain} {host}')
        
        domain_name, zone = self.detect_zone(domain)
        zone_class = self.detect_subclass(zone)
        self.child = zone_class()
        self.domain = Domain(domain_name, zone)

        logger.info(f'Zone Class {zone_class} {domain_name} {zone}')
        
        domain = f'{domain_name}.{zone}'

        if mock_from_txt_file is not None:
            f = open(mock_from_txt_file)
            raw = f.read()
            f.close()
        else:
            if host:
                command = ['whois', f'-h {host}', domain]
            else:
                command = ['whois', domain]

            p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            r = p.communicate()[0]
            raw = r.decode()
            
            if p.returncode != 0:
                error = f'WhoIs error {p.returncode} {raw}'
                logger.error(error)
                raise WhoIsCommandError(error)
            
        if self.child.is_free(raw):
            return None

        # raise any errors
        self.child.check_errors(raw)

        self.domain.is_free = False
        # parse raw data
        self.child.parse(raw, self)       

    def detect_zone(self, domain):
        logger.info(f'Detect zone {domain}')
        
        domain = domain.lower().strip()

        parts = domain.split('.')
        if parts[0].startswith('https://'):
            parts[0].replace('https://', '')
        elif parts[0].startswith('http://'):
            parts[0].replace('http://', '')

        domain = parts[0]
        zone = '.'.join(parts[1:])

        return domain, zone
    
    def detect_subclass(self, zone):
        
        logger.info(f'Detecting subclass for {zone} at {subclasses}')
        
        for cls in subclasses:
            logger.info(f'Searching zones for {cls} {cls.zones()}')
            if zone in cls.zones():
                return cls

        error = f'Zone not covered "{zone}"'
        logger.error(error)
        raise ZoneNotFoundError(error)



from datetime import datetime
import pytz
from whoare.whoare import WhoAre

tz = pytz.timezone('America/Argentina/Cordoba')


def test_data99():

    wa = WhoAre()    
    wa.load('data99.com.ar', mock_from_txt_file='whoare/zone_parsers/ar/sample_data99.txt')
    
    assert wa.domain.base_name == 'data99'
    assert wa.domain.zone == 'com.ar'
    assert not wa.domain.is_free
    assert wa.domain.registered == tz.localize(datetime(2010, 4, 12, 0, 0, 0), is_dst=True)
    assert wa.domain.changed == tz.localize(datetime(2020, 3, 24, 8, 26, 1, 899786), is_dst=True)
    assert wa.domain.expire == tz.localize(datetime(2021, 4, 12, 0, 0, 0), is_dst=True)
    
    assert wa.registrant.name == 'gomez pedro'
    assert wa.registrant.legal_uid == '20xxxxxxxx8'
    assert wa.registrant.created == tz.localize(datetime(2013, 8, 20, 0, 0, 0), is_dst=True)
    assert wa.registrant.changed == tz.localize(datetime(2020, 5, 4, 19, 34, 57, 928489), is_dst=True)
    
    assert wa.dnss[0].name == 'ns2.cluster311.com'
    assert wa.dnss[1].name == 'ns1.cluster311.com'


def test_free_01():

    wa = WhoAre()
    wa.load('domainfree.com.ar', mock_from_txt_file='whoare/zone_parsers/ar/sample_free.txt')
    assert wa.domain.is_free

def test_fernet():

    wa = WhoAre()
    wa.load('fernet.com.ar', mock_from_txt_file='whoare/zone_parsers/ar/sample_fernet.txt')
    
    assert wa.domain.base_name == 'fernet'
    assert wa.domain.zone == 'com.ar'
    assert not wa.domain.is_free
    assert wa.domain.registered == tz.localize(datetime(2020, 5, 7, 10, 44, 4, 210977), is_dst=True)
    assert wa.domain.changed == tz.localize(datetime(2020, 5, 7, 14, 34, 40, 828423), is_dst=True)
    assert wa.domain.expire == tz.localize(datetime(2021, 5, 7, 0, 0, 0), is_dst=True)

    assert wa.registrant.name == 'perez juan'
    assert wa.registrant.legal_uid == '20xxxxxxxx9'
    assert wa.registrant.created == tz.localize(datetime(2013, 8, 19, 0, 0, 0), is_dst=True)
    assert wa.registrant.changed == tz.localize(datetime(2020, 10, 11, 17, 37, 54, 831645), is_dst=True)
    
    assert wa.dnss[0].name == 'ns2.sedoparking.com'
    assert wa.dnss[1].name == 'ns1.sedoparking.com'


def test_nic():

    wa = WhoAre()
    wa.load('nic.ar', mock_from_txt_file='whoare/zone_parsers/ar/sample_nic.txt')
    
    assert wa.domain.base_name == 'nic'
    assert wa.domain.zone == 'ar'
    assert not wa.domain.is_free
    assert wa.domain.registered == tz.localize(datetime(1998, 5, 29, 0, 0, 0), is_dst=True)
    assert wa.domain.changed == tz.localize(datetime(2019, 10, 7, 16, 23, 59, 503535), is_dst=True)
    assert wa.domain.expire == tz.localize(datetime(2020, 5, 1, 0, 0, 0), is_dst=True)

    assert wa.registrant.name == 'nic.ar'
    assert wa.registrant.legal_uid == '99999999994'
    assert wa.registrant.created == tz.localize(datetime(2016, 9, 26, 12, 4, 48, 869673), is_dst=True)
    assert wa.registrant.changed == tz.localize(datetime(2016, 9, 26, 17, 39, 38, 373610), is_dst=True)
    
    assert wa.dnss[0].name == 'ns5.rdns.ar'
    assert wa.dnss[1].name == 'ns6.rdns.ar'
    assert wa.dnss[2].name == 'ns1.rdns.ar'
    assert wa.dnss[3].name == 'ns2.rdns.ar'
    assert wa.dnss[4].name == 'ns3.rdns.ar'


def test_fernet():

    wa = WhoAre()
    wa.load('fernet.com.ar', mock_from_txt_file='whoare/zone_parsers/ar/sample_fernet.txt')
    
    expected_dict = {
        "domain": {
            "base_name": 'fernet',
            "zone": 'com.ar',
            "is_free": False,
            "registered": wa.domain.registered,
            "changed": wa.domain.changed,
            "expire": wa.domain.expire
            },
        "registrant": {
            "name": wa.registrant.name,
            "legal_uid": wa.registrant.legal_uid,
            "created": wa.registrant.created,
            "changed": wa.registrant.changed
        },
        "dnss": ['ns2.sedoparking.com', 'ns1.sedoparking.com']
    }
    
    assert wa.as_dict() == expected_dict


# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='whoare',  
    # Versions should comply with PEP 440:
    # https://www.python.org/dev/peps/pep-0440/
    #
    # For a discussion on single-sourcing the version across setup.py and the
    # project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.1.25',  # Required
    description='A whois scraper',
    long_description=long_description,  # Optional
    long_description_content_type='text/markdown',  # Optional (see note above)

    # This should be a valid link to your project's main homepage.
    #
    # This field corresponds to the "Home-Page" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#home-page-optional
    url='https://github.com/avdata99/whoare',
    author='Anders Vazquez',
    author_email='andres@data99.com.ar',
    
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate you support Python 3. These classifiers are *not*
        # checked by 'pip install'. See instead 'python_requires' below.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
    ],

    keywords='whois, domain, nic.ar',
    
    packages=find_packages(exclude='tests'),  # Required
    python_requires='>=3.6, <4',

    install_requires=[
        'pytz',  # for localize dates
    ], 
    
    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    extras_require={  # Optional
        # 'dev': ['flake8']
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    package_data={  # Optional
        # 'sample': ['package_data.dat'],
    },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/distutils/setupscript.html#installing-additional-files
    #
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    data_files=[
        # ('my_data', ['data/data_file'])
        ],  # Optional

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    #
    # For example, the following would provide a command called `sample` which
    # executes the function `main` from this package when invoked:
    entry_points={  # Optional
        # 'console_scripts': [
        #     'sample=sample:main',
        # ],
    },

    # List additional URLs that are relevant to your project as a dict.
    #
    # This field corresponds to the "Project-URL" metadata fields:
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    #
    # Examples listed include a pattern for specifying where the package tracks
    # issues, where the source is hosted, where to say thanks to the package
    # maintainers, and where to support the project financially. The key is
    # what's used to render the link text on PyPI.
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/avdata99/whoare/issues',
        # 'Funding': 'https://donate.pypi.org',
        # 'Say Thanks!': 'http://saythanks.io/to/example',
        'Source': 'https://github.com/avdata99/whoare/',
    },
)



import pytz
from datetime import datetime
import logging
from whoare.exceptions import (TooManyQueriesError, ServiceUnavailableError, 
                               UnknownError, UnexpectedParseError,
                               UnexpectedDomainError)


logger = logging.getLogger(__name__)
tz = pytz.timezone('America/Argentina/Cordoba')


class WhoAr:

    @classmethod
    def zones(cls):
        return ['ar', 'com.ar', 'gob.ar', 'gov.ar', 'int.ar', 'mil.ar', 'org.ar']

    def is_free(self, raw):
        """ determine if domain is free """
        return "El dominio no se encuentra registrado" in raw

    def check_errors(self, raw):
        
        if "el servicio WHOIS de NIC Argentina se encuentra inactivo" in raw:
            raise ServiceUnavailableError()

        if "Excediste la cantidad permitida de consultas" in raw:
            raise TooManyQueriesError()

        if "fgets: Conexión reinicializada por la máquina remota" in raw:
            raise UnknownError()
        

    def parse(self, raw, parent):
        from whoare.base import Registrant, DNS
        logger.debug(f'Parsing {raw}')
        lines = raw.split('\n')

        real_lines = []
        for line in lines:
            if line.startswith('%') or line == '':
                continue
            real_lines.append(line)

        logger.debug(f'Parsing cleaned {real_lines}')
        
        # ==========================================
        field, value = self._parse_line(real_lines[0])
        if field != 'domain':
            raise UnexpectedParseError(f'Field {field} is not "domain"')
        
        if value != parent.domain.full_name():
            raise UnexpectedDomainError(f'Unexpected domain {value} != {parent.domain.full_name()}')
        
        # ==========================================
        field, value = self._parse_line(real_lines[1])
        if field != 'registrant':
            raise UnexpectedParseError(f'Field {field} is not "registrant"')
        
        registrant_uid = value

        # ignore line 2, registrar

        # ==========================================
        field, value = self._parse_line(real_lines[3])
        if field != 'registered':
            raise UnexpectedParseError(f'Field {field} is not "registered"')
        
        parent.domain.registered = self._get_nic_date(value)
        
        # ==========================================
        field, value = self._parse_line(real_lines[4])
        if field != 'changed':
            raise UnexpectedParseError(f'Field {field} is not "changed"')

        parent.domain.changed = self._get_nic_date(value)
        
        # ==========================================
        field, value = self._parse_line(real_lines[5])
        if field != 'expire':
            raise UnexpectedParseError(f'Field {field} is not "expire"')

        parent.domain.expire = self._get_nic_date(value)
        
        # ==========================================
        field, value = self._parse_line(real_lines[6])
        if field != 'contact':
            raise UnexpectedParseError(f'Field {field} is not "contact"')

        if value != registrant_uid:
            raise UnexpectedParseError(f'Legal UID diff {value} != {registrant_uid}')

        # ==========================================
        field, value = self._parse_line(real_lines[7])
        if field != 'name':
            raise UnexpectedParseError(f'Field {field} is not "name"')

        parent.registrant = Registrant(name=value, legal_uid=registrant_uid)

        # ignore line, registrar

        # ==========================================
        field, value = self._parse_line(real_lines[9])
        if field != 'created':
            raise UnexpectedParseError(f'Field {field} is not "created"')
        
        parent.registrant.created = self._get_nic_date(value)

        # ==========================================
        field, value = self._parse_line(real_lines[10])
        if field != 'changed':
            raise UnexpectedParseError(f'Field {field} is not "changed"')
        
        parent.registrant.changed = self._get_nic_date(value)

        # ==========================================
        field, value = self._parse_line(real_lines[11])
        n = 11
        while field == "nserver":
            parts = value.split()
            ns = parts[0]
            logger.info(f'DNS found {ns}')
            parent.dnss.append(DNS(name=ns))
            n += 1
            field, value = self._parse_line(real_lines[n])
        
    def _parse_line(self, line):
        parts = line.split(':')
        field = parts[0].lower().strip()
        value = ':'.join(parts[1:]).lower().strip()

        return field, value

    def _get_nic_date(self, date_str):
        """ nic si es muy nuevo ya devuelve milisengundos """
        
        try:
            res = datetime.strptime(date_str, '%Y-%m-%d %H:%M:%S.%f')
        except ValueError:
            res = datetime.strptime(date_str, '%Y-%m-%d %H:%M:%S')

        res = tz.localize(res, is_dst=True)        
        logger.debug(f'_get_nic_date {date_str} {res}')
        return res
